
// Author - Amit Sengupta
// email  - hilogicsys@gmail.com

/*************************************************************************
                 user_main.c: Main entry-point 
/************************************************************************/

#include "ets_sys.h"
#include "osapi.h"
#include "gpio.h"
#include "os_type.h"    
#include "ip_addr.h"
#include "espconn.h"
#include "user_interface.h"

#include "eagle_soc.h"

#include "espmissingincludes.h"
#include "UART/uart.h"
#include "SPI/spi_interface.h"
#include "Mp3.h"

#define USER_TASK_PRIORITY          0
#define USER_TASK_QUEUE_LENGTH      1

// Change the below values to suit your own network.
#define SSID   "Tenda_2CF970"
#define PASSWD "80518051"

#define BUFF_MAX    16384

#define F_UDPVALID  1
#define F_SPIVALID  2
#define F_UDPBF1    4
#define F_SPIBF1    8
#define F_SPIREQ    16

LOCAL void ICACHE_FLASH_ATTR Set_Timer(uint16_t interval);
LOCAL void ICACHE_FLASH_ATTR UDPTX_Data(uint8_t *data, uint16_t len, uint32_t ip_addr); 
LOCAL void ICACHE_FLASH_ATTR HSPI_Init(void);
LOCAL void ICACHE_FLASH_ATTR HSPI_Tx(uint8_t* Ptr_Buff);
LOCAL void ICACHE_FLASH_ATTR gpio_intr_handler(void);
LOCAL void ICACHE_FLASH_ATTR Tx_SPIBufferBeep(void);
LOCAL void ICACHE_FLASH_ATTR read_rom_uint32(uint8* Dest,const uint32* Src);

os_event_t userTaskQueue[USER_TASK_QUEUE_LENGTH];
// Timer used to determine when the LED is to be turned on/off.
LOCAL os_timer_t UDPRxTimer;

// The current state of the LED's output.
LOCAL uint8_t led_state = 0;

// Structure holding the UDP connection information.
LOCAL struct espconn UDPRX_Conn;
// Connection used to transmit UDP packets.
LOCAL struct espconn UDPTX_Conn;

// UDP specific protocol structure.
LOCAL esp_udp UDPRX_Proto;
// UDP specific procotol structure used for transmitting UDP packets.
LOCAL esp_udp UDPTX_Proto;

// Remote Computer IP Address
LOCAL uint32_t Remote_IPADD = 0;
// Flag variable 
LOCAL uint8_t FLAGS = 0;

// Data Buffers
LOCAL uint8_t Data_BUFF1[BUFF_MAX]; 
LOCAL uint8_t Data_BUFF2[BUFF_MAX]; 
LOCAL uint8_t* Ptr_BUFFUDP;
LOCAL uint8_t* Ptr_BUFFSPI;
LOCAL uint16_t UDPRxTOT_Len;


//  User function which will be added to the task queue.
static void ICACHE_FLASH_ATTR UserTask(os_event_t *events)
{
    system_os_post(USER_TASK_PRIORITY, 0, 0 );
    if((FLAGS & F_SPIREQ) != F_SPIREQ)return;

    uint16_t i,j;
    if(Remote_IPADD == 0)
    {
       FLAGS = FLAGS & ~F_SPIREQ;
       Tx_SPIBufferBeep();
       return;
    }
  
    if((FLAGS & F_SPIVALID) == F_SPIVALID)
    {
        FLAGS = FLAGS & ~F_SPIREQ;
        //os_printf("\nUDP DATA REQ (N)");
        FLAGS = FLAGS & ~F_UDPVALID;         // Clear UDP Data Valid Flag
        FLAGS = FLAGS & ~F_SPIVALID;         // Clear SPI Data Valid Flag
        UDPTX_Data("N",1,Remote_IPADD);
        Set_Timer(500);
        if((FLAGS & F_SPIBF1) == F_SPIBF1 )
        {
          Ptr_BUFFSPI = Data_BUFF1;
        }
        else 
        {
          Ptr_BUFFSPI = Data_BUFF2;
        }
        FLAGS = FLAGS & ~F_SPIREQ;
        
        //os_printf("\nSPI Tx - S");
        for(i = 0; i < 256; i++)
        { 
          HSPI_Tx(Ptr_BUFFSPI);
          Ptr_BUFFSPI = Ptr_BUFFSPI + 64;
        }
        //os_printf("\nSPI Tx - E");
    }
}

LOCAL void ICACHE_FLASH_ATTR Tx_SPIBufferBeep()
{
    uint16_t i,j;
    uint16_t ind_MP3 = 0;

    for(i = 0; i < 256;i++)
    {
       Ptr_BUFFSPI = Data_BUFF1; 
       for(j = 0; j < 16; j++)
       {
          read_rom_uint32(Ptr_BUFFSPI,(uint32_t*)(FILE_MP3)+ind_MP3);
          Ptr_BUFFSPI = Ptr_BUFFSPI+4;
          ind_MP3++;
       }
       Ptr_BUFFSPI = Data_BUFF1;  
       HSPI_Tx(Ptr_BUFFSPI);
    }
}

LOCAL void ICACHE_FLASH_ATTR read_rom_uint32(uint8* Dest,const uint32* Src)
{
    Dest[0] = Src[0];
    Dest[1] = Src[0] >> 8;
    Dest[2] = Src[0] >> 16;
    Dest[3] = Src[0] >> 24;
}

/*************************************************************************
        Sends a UDP packet to the given IP address.
*************************************************************************/
LOCAL void ICACHE_FLASH_ATTR UDPTX_Data(uint8_t *data, uint16_t len, uint32_t ip_addr) 
{
    // Set the destination IP address and port.
    os_memcpy(UDPTX_Proto.remote_ip, &ip_addr, 4);
    espconn_send(&UDPTX_Conn, data, len);
}

/*************************************************************************
                        Timer Callback
*************************************************************************/
LOCAL void ICACHE_FLASH_ATTR Timer_Cb(void *arg) 
{
    FLAGS = FLAGS & ~F_UDPVALID;       // Clear UDP Data Valid Flag
    UDPTX_Data("N",1,Remote_IPADD);
    Tx_SPIBufferBeep();
}

/*************************************************************************
                    Sets the Interval of the Timer
*************************************************************************/
LOCAL void ICACHE_FLASH_ATTR Set_Timer(uint16_t interval) 
{
    // Start a timer for the flashing of the LED on GPIO 4, running continuously.
    os_timer_disarm(&UDPRxTimer);
    os_timer_setfn(&UDPRxTimer, (os_timer_func_t *)Timer_Cb, (void *)0);
    os_timer_arm(&UDPRxTimer, interval, 1);
}

/*************************************************************************
        Call-back for changes in the WIFi connection's state.
*************************************************************************/
LOCAL void ICACHE_FLASH_ATTR wifi_event_cb(System_Event_t *event) 
{
    struct ip_info info;
    ip_addr_t LocalIP;

    // To determine what actually happened, we need to look at the event.
    switch (event->event) 
    {
        case EVENT_STAMODE_CONNECTED: 
        {
            // We are connected as a station, but we don't have an IP address yet.
            char ssid[33];
            uint8_t len = event->event_info.connected.ssid_len;
            if (len > 32)len = 32;
            strncpy(ssid, event->event_info.connected.ssid, len + 1);
            os_printf("Received EVENT_STAMODE_CONNECTED. "
                      "SSID = %s, BSSID = "MACSTR", channel = %d.\n",
                      ssid, MAC2STR(event->event_info.connected.bssid), event->event_info.connected.channel);
            //Set_Timer(500);
            break;
        }
        case EVENT_STAMODE_DISCONNECTED: 
        {
            // We have been disconnected as a station.
            char ssid[33];
            uint8_t len = event->event_info.connected.ssid_len;
            if (len > 32)len = 32;
            strncpy(ssid, event->event_info.connected.ssid, len + 1);
            os_printf("Received EVENT_STAMODE_DISCONNECTED. "
                      "SSID = %s, BSSID = "MACSTR", channel = %d.\n",
                      ssid, MAC2STR(event->event_info.disconnected.bssid), event->event_info.disconnected.reason);
            //Set_Timer(100);
            Remote_IPADD = 0;
            break;
        }
        case EVENT_STAMODE_GOT_IP:
             
             // We have an IP address, ready to run. Return the IP address, too.
             os_printf("Received EVENT_STAMODE_GOT_IP. IP = "IPSTR", mask = "IPSTR", gateway = "IPSTR"\n", 
                      IP2STR(&event->event_info.got_ip.ip.addr), 
                      IP2STR(&event->event_info.got_ip.mask.addr),
                      IP2STR(&event->event_info.got_ip.gw));
             //Set_Timer(4000);
             gpio_output_set(0, BIT2, BIT2, 0);
             break;
        case EVENT_STAMODE_DHCP_TIMEOUT:
            // We couldn't get an IP address via DHCP, so we'll have to try re-connecting.
            os_printf("Received EVENT_STAMODE_DHCP_TIMEOUT.\n");
            wifi_station_disconnect();
            wifi_station_connect();
            //Set_Timer(100);
            break;
    }
}

/*************************************************************************
               Handles the receiving UDP Datagram
*************************************************************************/
LOCAL void ICACHE_FLASH_ATTR UDPRec_Cb(void *arg, char *data, uint16_t len) 
{
    uint16_t i;
    
    //os_printf("\nReceived data from "IPSTR, IP2STR(&Remote_IPADD));
    if(Remote_IPADD == 0 )
    {
      remot_info* premot = NULL;
      ip_addr_t RemoteIP_ADD;
      uint8_t* RemoteIP_AR = NULL;

      espconn_get_connection_info(&UDPRX_Conn,&premot,0);
      RemoteIP_AR = premot->remote_ip;
      IP4_ADDR(&RemoteIP_ADD, RemoteIP_AR[0], RemoteIP_AR[1], RemoteIP_AR[2], RemoteIP_AR[3]);
      Remote_IPADD = RemoteIP_ADD.addr;
      UDPTX_Proto.remote_port = 1234;//premot->remote_port; 
      UDPTX_Data("N",1,Remote_IPADD);
      FLAGS = FLAGS | F_SPIVALID;      // Set SPI Data Valid Flag
      return;
    }
    
    os_memcpy(Ptr_BUFFUDP,data,len);
    UDPRxTOT_Len = UDPRxTOT_Len + len;
    Ptr_BUFFUDP = Ptr_BUFFUDP + len;  
 
    //os_printf("\nUDP TOT = %d ",UDPRxTOT_Len);

    if(UDPRxTOT_Len >= BUFF_MAX)
    {
       os_timer_disarm(&UDPRxTimer);
       UDPRxTOT_Len = 0;
       if((FLAGS & F_UDPBF1) == F_UDPBF1 )
       {
          Ptr_BUFFUDP = Data_BUFF2;
          FLAGS = FLAGS & ~F_UDPBF1; 
          FLAGS = FLAGS | F_SPIBF1; 
          //os_printf("Ptr_BUFFUDP = Data_BUFF2\n");
       }
       else 
       {
          Ptr_BUFFUDP = Data_BUFF1;
          FLAGS = FLAGS | F_UDPBF1; 
          FLAGS = FLAGS & ~F_SPIBF1; 
          //os_printf("Ptr_BUFFUDP = Data_BUFF1\n");
       }
       FLAGS = FLAGS | F_UDPVALID;  // Set UDP Data Valid Flag
       FLAGS = FLAGS | F_SPIVALID;  // Set SPI Data Valid Flag
       //os_printf("\nUDP Rx");
       return;
    }
    led_state = !led_state;
    GPIO_OUTPUT_SET(2, led_state);
}

/*************************************************************************
              Sets up the WiFi interface on the ESP-8266.
*************************************************************************/
LOCAL void ICACHE_FLASH_ATTR wifi_init() 
{

    // Set station mode - Connect to a WiFi router.
    wifi_set_opmode_current(STATION_MODE);
    wifi_set_broadcast_if(STATION_MODE);

    // Set up the network name and password.
    struct station_config sc;
    strncpy(sc.ssid, SSID, 32);
    strncpy(sc.password, PASSWD, 64);
    wifi_station_set_config(&sc);
    wifi_station_dhcpc_start();

    // Set up the call back for the status of the WiFi.
    wifi_set_event_handler_cb(wifi_event_cb);

    // Setup the UDP Client.
    UDPRX_Proto.local_port = 1900;
    UDPRX_Conn.type        = ESPCONN_UDP;
    UDPRX_Conn.state       = ESPCONN_NONE;
    UDPRX_Conn.proto.udp   = &UDPRX_Proto;

    espconn_create(&UDPRX_Conn);
    espconn_regist_recvcb(&UDPRX_Conn, UDPRec_Cb);
    
    // Setup the UDP Server
    UDPTX_Conn.type      = ESPCONN_UDP;
    UDPTX_Conn.state     = ESPCONN_NONE;
    UDPTX_Conn.proto.udp = &UDPTX_Proto;

    espconn_create(&UDPTX_Conn);
}

void ICACHE_FLASH_ATTR HSPI_Tx(uint8_t* Ptr_Buff)
{
    uint32_t addr = 0;
    SpiData spiData;

    spiData.cmd = MASTER_WRITE_DATA_TO_SLAVE_CMD;
    spiData.cmdLen = 0;
    spiData.addr = &addr;
    spiData.addrLen = 0;
    spiData.data = (uint32_t*)Ptr_Buff; 
    spiData.dataLen = 64;
    SPIMasterSendData(SpiNum_HSPI, &spiData);
}

/*************************************************************************
  Receives Data from the serial port. We're ignoring characters here.
*************************************************************************/
void ICACHE_FLASH_ATTR uart_rx_task(os_event_t *events) 
{
    uint8_t Ch;
    if (events->sig == 0) 
    {
        // Sig 0 is a normal receive. Get how many bytes have been received.
        uint8_t rx_len = (READ_PERI_REG(UART_STATUS(UART0)) >> UART_RXFIFO_CNT_S) & UART_RXFIFO_CNT;

        // Read the characters from the FIFO, and discard them.
        for (uint8_t ii=0; ii < rx_len; ii++) 
        {
            Ch = READ_PERI_REG(UART_FIFO(UART0)) & 0xFF;
            uart_tx_one_char(UART0, Ch);
        }
        // Clear the interrupt condition flags and re-enable the receive interrupt.
        WRITE_PERI_REG(UART_INT_CLR(UART0), UART_RXFIFO_FULL_INT_CLR | UART_RXFIFO_TOUT_INT_CLR);
        uart_rx_intr_enable(UART0);
    }
}

// HSPI Initialize
LOCAL void ICACHE_FLASH_ATTR HSPI_Init()
{
    SpiAttr hSpiAttr;
    hSpiAttr.bitOrder = SpiBitOrder_MSBFirst;
    hSpiAttr.speed    = SpiSpeed_10MHz;
    hSpiAttr.mode     = SpiMode_Master;
    hSpiAttr.subMode  = SpiSubMode_0;

    // Init HSPI GPIO
    WRITE_PERI_REG(PERIPHS_IO_MUX, 0x105);

    //PIN_FUNC_SELECT(PERIPHS_IO_MUX_MTDI_U, 2); // GPIO12 - MISO
    PIN_FUNC_SELECT(PERIPHS_IO_MUX_MTCK_U, 2); // GPIO13 - MOSI
    PIN_FUNC_SELECT(PERIPHS_IO_MUX_MTMS_U, 2); // GPIO14 - CLOCK
    PIN_FUNC_SELECT(PERIPHS_IO_MUX_MTDO_U, 2); // GPIO15 - CS/SS

    SPIInit(SpiNum_HSPI, &hSpiAttr);
}

// GPIO12 Interrupt handler
LOCAL void ICACHE_FLASH_ATTR gpio_intr_handler()
{
   // clear gpio status. 
   uint32 gpio_status = GPIO_REG_READ(GPIO_STATUS_ADDRESS);
   // if the interrupt was by GPIO12
    if (gpio_status & BIT12)
    {
        // Disable interrupt for GPIO12
        gpio_pin_intr_state_set(GPIO_ID_PIN(12), GPIO_PIN_INTR_DISABLE);
        
        //os_printf("\nGPIO12");
        FLAGS = FLAGS | F_SPIREQ;
        
        //clear interrupt status for GPIO12
        GPIO_REG_WRITE(GPIO_STATUS_W1TC_ADDRESS, gpio_status & BIT12);
        // Reactivate interrupts for GPIO12
        gpio_pin_intr_state_set(GPIO_ID_PIN(12), GPIO_PIN_INTR_NEGEDGE);
    }
}
   
/*************************************************************************
                        Entry point for the program.
*************************************************************************/
void user_init(void) 
{

    // Initialise the serial port.
    uart_init(BIT_RATE_115200, BIT_RATE_115200);

    // Initialise all GPIOs.
    gpio_init();

    // GPIO 2 is an output, start with it low (off).
    PIN_FUNC_SELECT(PERIPHS_IO_MUX_GPIO2_U, FUNC_GPIO2);
    
    Ptr_BUFFUDP = Data_BUFF1;
    Ptr_BUFFSPI = Data_BUFF1; 
    UDPRxTOT_Len = 0;
    FLAGS = FLAGS | F_UDPBF1;     // UDP Buffer = Data_BUFF1
    FLAGS = FLAGS | F_SPIBF1;     // SPI Buffer = Data_BUFF1
    FLAGS = FLAGS & ~F_UDPVALID;  // Clear UDP Data Valid Flag
    FLAGS = FLAGS & ~F_SPIVALID;  // Clear SPI Data Valid Flag
    FLAGS = FLAGS & ~F_SPIREQ;    // Clear SPI Data Req Flag

    os_printf("*** RESET ***\n");
    // Start the network.
    wifi_init();
    
    // Setup GPIO12 PIN Interrupt
    PIN_FUNC_SELECT(PERIPHS_IO_MUX_MTDI_U, FUNC_GPIO12);
    GPIO_DIS_OUTPUT(12);
    PIN_PULLUP_DIS(PERIPHS_IO_MUX_MTDI_U);
    PIN_PULLUP_EN(PERIPHS_IO_MUX_MTDI_U);
    ETS_GPIO_INTR_DISABLE();
    ETS_GPIO_INTR_ATTACH(gpio_intr_handler,NULL);
    gpio_pin_intr_state_set(GPIO_ID_PIN(12),GPIO_PIN_INTR_NEGEDGE);
    ETS_GPIO_INTR_ENABLE();

    // Initialise the HSPI port.
    HSPI_Init();

    system_os_task(UserTask, USER_TASK_PRIORITY, userTaskQueue, USER_TASK_QUEUE_LENGTH);
    system_os_post(USER_TASK_PRIORITY, 0, 0 );
}
